package com.sun.xml.internal.ws.client.sei;

import com.sun.xml.internal.ws.spi.db.PropertyAccessor;
import com.sun.xml.internal.ws.spi.db.DatabindingException;
import java.util.List;
import com.sun.xml.internal.ws.spi.db.WrapperComposite;
import com.sun.xml.internal.ws.model.WrapperParameter;
import com.sun.xml.internal.ws.api.message.Header;
import javax.xml.soap.SOAPFault;
import javax.xml.soap.SOAPException;
import com.sun.xml.internal.ws.resources.ServerMessages;
import javax.xml.ws.soap.SOAPFaultException;
import com.sun.xml.internal.ws.api.model.ParameterBinding;
import com.sun.xml.internal.ws.api.SOAPVersion;
import java.io.IOException;
import javax.imageio.ImageIO;
import com.sun.xml.internal.ws.encoding.DataHandlerDataSource;
import com.sun.xml.internal.ws.encoding.StringDataContentHandler;
import java.util.Iterator;
import java.io.InputStream;
import java.awt.Image;
import javax.xml.transform.Source;
import javax.activation.DataHandler;
import com.sun.xml.internal.ws.model.ParameterImpl;
import java.util.Collection;
import java.util.ArrayList;
import com.sun.xml.internal.ws.spi.db.RepeatedElementBridge;
import com.sun.xml.internal.ws.message.AttachmentUnmarshallerImpl;
import com.sun.xml.internal.ws.api.message.AttachmentSet;
import com.sun.xml.internal.ws.spi.db.XMLBridge;
import java.util.HashMap;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import com.sun.xml.internal.ws.api.message.Attachment;
import java.lang.reflect.Type;
import javax.xml.stream.XMLStreamReader;
import javax.xml.ws.WebServiceException;
import com.sun.xml.internal.ws.api.streaming.XMLStreamReaderFactory;
import com.sun.xml.internal.ws.streaming.XMLStreamReaderUtil;
import javax.xml.stream.XMLStreamException;
import javax.xml.bind.JAXBException;
import com.sun.xml.internal.ws.api.message.Message;
import javax.xml.namespace.QName;
import java.util.Map;

public abstract class ResponseBuilder
{
    protected Map<QName, ResponseBuilder.WrappedPartBuilder> wrappedParts;
    protected QName wrapperName;
    public static final ResponseBuilder NONE;
    private static final Map<Class, Object> primitiveUninitializedValues;
    
    public ResponseBuilder() {
        this.wrappedParts = null;
    }
    
    public abstract Object readResponse(Message p0, Object[] p1) throws JAXBException, XMLStreamException;
    
    protected Object readWrappedResponse(Message msg, Object[] args) throws JAXBException, XMLStreamException {
        Object retVal = null;
        if (!msg.hasPayload()) {
            throw new WebServiceException("No payload. Expecting payload with " + this.wrapperName + " element");
        }
        XMLStreamReader reader = msg.readPayload();
        XMLStreamReaderUtil.verifyTag(reader, this.wrapperName);
        reader.nextTag();
        while (reader.getEventType() == 1) {
            ResponseBuilder.WrappedPartBuilder part = (ResponseBuilder.WrappedPartBuilder)this.wrappedParts.get(reader.getName());
            if (part == null) {
                XMLStreamReaderUtil.skipElement(reader);
                reader.nextTag();
            }
            else {
                Object o = part.readResponse(args, reader, msg.getAttachments());
                if (o != null) {
                    assert retVal == null;
                    retVal = o;
                }
            }
            if (reader.getEventType() != 1 && reader.getEventType() != 2) {
                XMLStreamReaderUtil.nextElementContent(reader);
            }
            else {
                continue;
            }
        }
        reader.close();
        XMLStreamReaderFactory.recycle(reader);
        return retVal;
    }
    
    public static Object getVMUninitializedValue(Type type) {
        return ResponseBuilder.primitiveUninitializedValues.get(type);
    }
    
    public static final String getWSDLPartName(Attachment att) {
        String cId = att.getContentId();
        int index = cId.lastIndexOf(64, cId.length());
        if (index == -1) {
            return null;
        }
        String localPart = cId.substring(0, index);
        index = localPart.lastIndexOf(61, localPart.length());
        if (index == -1) {
            return null;
        }
        try {
            return URLDecoder.decode(localPart.substring(0, index), "UTF-8");
        }
        catch (UnsupportedEncodingException e) {
            throw new WebServiceException(e);
        }
    }
    
    private static boolean isXMLMimeType(String mimeType) {
        return mimeType.equals("text/xml") || mimeType.equals("application/xml");
    }
    
    static boolean access$100(String x0) {
        return isXMLMimeType(x0);
    }
    
    static {
        NONE = new ResponseBuilder.None();
        primitiveUninitializedValues = new HashMap();
        Map<Class, Object> m = ResponseBuilder.primitiveUninitializedValues;
        m.put(Integer.TYPE, Integer.valueOf(0));
        m.put(Character.TYPE, Character.valueOf('\0'));
        m.put(Byte.TYPE, Byte.valueOf((byte)0));
        m.put(Short.TYPE, Short.valueOf((short)0));
        m.put(Long.TYPE, Long.valueOf(0L));
        m.put(Float.TYPE, Float.valueOf(0.0f));
        m.put(Double.TYPE, Double.valueOf(0.0));
    }
    
    static final class WrappedPartBuilder
    {
        private final XMLBridge bridge;
        private final ValueSetter setter;
        
        public WrappedPartBuilder(XMLBridge bridge, ValueSetter setter) {
            this.bridge = bridge;
            this.setter = setter;
        }
        
        final Object readResponse(Object[] args, XMLStreamReader r, AttachmentSet att) throws JAXBException {
            AttachmentUnmarshallerImpl au = (att != null) ? new AttachmentUnmarshallerImpl(att) : null;
            Object obj;
            if ((this.bridge instanceof RepeatedElementBridge)) {
                RepeatedElementBridge rbridge = (RepeatedElementBridge)this.bridge;
                ArrayList<Object> list = new ArrayList();
                QName name = r.getName();
                while (r.getEventType() == 1 && name.equals(r.getName())) {
                    list.add(rbridge.unmarshal(r, au));
                    XMLStreamReaderUtil.toNextTag(r, name);
                }
                obj = rbridge.collectionHandler().convert(list);
            }
            else {
                obj = this.bridge.unmarshal(r, au);
            }
            return this.setter.put(obj, args);
        }
    }
    
    static final class None extends ResponseBuilder
    {
        private None() {
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) {
            msg.consume();
            return null;
        }
    }
    
    public static final class NullSetter extends ResponseBuilder
    {
        private final ValueSetter setter;
        private final Object nullValue;
        
        public NullSetter(ValueSetter setter, Object nullValue) {
            assert setter != null;
            this.nullValue = nullValue;
            this.setter = setter;
            return;
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) {
            return this.setter.put(this.nullValue, args);
        }
    }
    
    public static final class Composite extends ResponseBuilder
    {
        private final ResponseBuilder[] builders;
        
        public Composite(ResponseBuilder... builders) {
            this.builders = builders;
        }
        
        public Composite(Collection<? extends ResponseBuilder> builders) {
            this((ResponseBuilder[])builders.toArray(new ResponseBuilder[builders.size()]));
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) throws JAXBException, XMLStreamException {
            Object retVal = null;
            for (ResponseBuilder builder : this.builders) {
                Object r = builder.readResponse(msg, args);
                if (r != null) {
                    assert retVal == null;
                    retVal = r;
                }
            }
            return retVal;
        }
    }
    
    public abstract static class AttachmentBuilder extends ResponseBuilder
    {
        protected final ValueSetter setter;
        protected final ParameterImpl param;
        private final String pname;
        private final String pname1;
        
        AttachmentBuilder(ParameterImpl param, ValueSetter setter) {
            this.setter = setter;
            this.param = param;
            this.pname = param.getPartName();
            this.pname1 = "<" + this.pname;
        }
        
        public static ResponseBuilder createAttachmentBuilder(ParameterImpl param, ValueSetter setter) {
            Class<Object> type = (Class)param.getTypeInfo().type;
            if (DataHandler.class.isAssignableFrom(type)) {
                return new ResponseBuilder.DataHandlerBuilder(param, setter);
            }
            if (byte[].class == type) {
                return new ResponseBuilder.ByteArrayBuilder(param, setter);
            }
            if (Source.class.isAssignableFrom(type)) {
                return new ResponseBuilder.SourceBuilder(param, setter);
            }
            if (Image.class.isAssignableFrom(type)) {
                return new ResponseBuilder.ImageBuilder(param, setter);
            }
            if (InputStream.class == type) {
                return new ResponseBuilder.InputStreamBuilder(param, setter);
            }
            if (ResponseBuilder.access$100(param.getBinding().getMimeType())) {
                return new ResponseBuilder.JAXBBuilder(param, setter);
            }
            if (String.class.isAssignableFrom(type)) {
                return new ResponseBuilder.StringBuilder(param, setter);
            }
            throw new UnsupportedOperationException("Unexpected Attachment type =" + type);
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) throws JAXBException, XMLStreamException {
            for (Attachment att : msg.getAttachments()) {
                String part = ResponseBuilder.getWSDLPartName(att);
                if (part == null) {
                    continue;
                }
                if (part.equals(this.pname) || part.equals(this.pname1)) {
                    return this.mapAttachment(att, args);
                }
                continue;
            }
            return null;
        }
        
        abstract Object mapAttachment(Attachment p0, Object[] p1) throws JAXBException;
    }
    
    private static final class DataHandlerBuilder extends ResponseBuilder.AttachmentBuilder
    {
        DataHandlerBuilder(ParameterImpl param, ValueSetter setter) {
            super(param, setter);
        }
        
        @Override
        Object mapAttachment(Attachment att, Object[] args) {
            return this.setter.put(att.asDataHandler(), args);
        }
    }
    
    private static final class StringBuilder extends ResponseBuilder.AttachmentBuilder
    {
        StringBuilder(ParameterImpl param, ValueSetter setter) {
            super(param, setter);
        }
        
        @Override
        Object mapAttachment(Attachment att, Object[] args) {
            att.getContentType();
            StringDataContentHandler sdh = new StringDataContentHandler();
            try {
                String str = (String)sdh.getContent(new DataHandlerDataSource(att.asDataHandler()));
                return this.setter.put(str, args);
            }
            catch (Exception e) {
                throw new WebServiceException(e);
            }
        }
    }
    
    private static final class ByteArrayBuilder extends ResponseBuilder.AttachmentBuilder
    {
        ByteArrayBuilder(ParameterImpl param, ValueSetter setter) {
            super(param, setter);
        }
        
        @Override
        Object mapAttachment(Attachment att, Object[] args) {
            return this.setter.put(att.asByteArray(), args);
        }
    }
    
    private static final class SourceBuilder extends ResponseBuilder.AttachmentBuilder
    {
        SourceBuilder(ParameterImpl param, ValueSetter setter) {
            super(param, setter);
        }
        
        @Override
        Object mapAttachment(Attachment att, Object[] args) {
            return this.setter.put(att.asSource(), args);
        }
    }
    
    private static final class ImageBuilder extends ResponseBuilder.AttachmentBuilder
    {
        ImageBuilder(ParameterImpl param, ValueSetter setter) {
            super(param, setter);
        }
        
        @Override
        Object mapAttachment(Attachment att, Object[] args) {
            InputStream is = null;
            Image image;
            try {
                is = att.asInputStream();
                image = ImageIO.read(is);
            }
            catch (IOException ioe) {
                throw new WebServiceException(ioe);
            }
            finally {
                if (is != null) {
                    try {
                        is.close();
                    }
                    catch (IOException ioe2) {
                        throw new WebServiceException(ioe2);
                    }
                }
            }
            return this.setter.put(image, args);
        }
    }
    
    private static final class InputStreamBuilder extends ResponseBuilder.AttachmentBuilder
    {
        InputStreamBuilder(ParameterImpl param, ValueSetter setter) {
            super(param, setter);
        }
        
        @Override
        Object mapAttachment(Attachment att, Object[] args) {
            return this.setter.put(att.asInputStream(), args);
        }
    }
    
    private static final class JAXBBuilder extends ResponseBuilder.AttachmentBuilder
    {
        JAXBBuilder(ParameterImpl param, ValueSetter setter) {
            super(param, setter);
        }
        
        @Override
        Object mapAttachment(Attachment att, Object[] args) throws JAXBException {
            Object obj = this.param.getXMLBridge().unmarshal(att.asInputStream());
            return this.setter.put(obj, args);
        }
    }
    
    public static final class Header extends ResponseBuilder
    {
        private final XMLBridge<?> bridge;
        private final ValueSetter setter;
        private final QName headerName;
        private final SOAPVersion soapVersion;
        
        public Header(SOAPVersion soapVersion, QName name, XMLBridge<?> bridge, ValueSetter setter) {
            this.soapVersion = soapVersion;
            this.headerName = name;
            this.bridge = bridge;
            this.setter = setter;
        }
        
        public Header(SOAPVersion soapVersion, ParameterImpl param, ValueSetter setter) {
            this(soapVersion, param.getTypeInfo().tagName, param.getXMLBridge(), setter);
            assert param.getOutBinding() == ParameterBinding.HEADER;
            return;
        }
        
        private SOAPFaultException createDuplicateHeaderException() {
            try {
                SOAPFault fault = this.soapVersion.getSOAPFactory().createFault();
                fault.setFaultCode(this.soapVersion.faultCodeServer);
                fault.setFaultString(ServerMessages.DUPLICATE_PORT_KNOWN_HEADER(this.headerName));
                return new SOAPFaultException(fault);
            }
            catch (SOAPException e) {
                throw new WebServiceException(e);
            }
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) throws JAXBException {
            com.sun.xml.internal.ws.api.message.Header header = null;
            Iterator<com.sun.xml.internal.ws.api.message.Header> it = msg.getHeaders().getHeaders(this.headerName, true);
            if (it.hasNext()) {
                header = (com.sun.xml.internal.ws.api.message.Header)it.next();
                if (it.hasNext()) {
                    throw this.createDuplicateHeaderException();
                }
            }
            if (header != null) {
                return this.setter.put(header.readAsJAXB(this.bridge), args);
            }
            return null;
        }
    }
    
    public static final class Body extends ResponseBuilder
    {
        private final XMLBridge<?> bridge;
        private final ValueSetter setter;
        
        public Body(XMLBridge<?> bridge, ValueSetter setter) {
            this.bridge = bridge;
            this.setter = setter;
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) throws JAXBException {
            return this.setter.put(msg.readPayloadAsJAXB(this.bridge), args);
        }
    }
    
    public static final class DocLit extends ResponseBuilder
    {
        private final ResponseBuilder.DocLit.PartBuilder[] parts;
        private final XMLBridge wrapper;
        private boolean dynamicWrapper;
        
        public DocLit(WrapperParameter wp, ValueSetterFactory setterFactory) {
            this.wrapperName = wp.getName();
            this.wrapper = wp.getXMLBridge();
            Class<Object> wrapperType = (Class)this.wrapper.getTypeInfo().type;
            this.dynamicWrapper = WrapperComposite.class.equals(wrapperType);
            List<ResponseBuilder.DocLit.PartBuilder> tempParts = new ArrayList();
            List<ParameterImpl> children = wp.getWrapperChildren();
            for (ParameterImpl p : children) {
                if (p.isIN()) {
                    continue;
                }
                QName name = p.getName();
                if (this.dynamicWrapper) {
                    if (this.wrappedParts == null) {
                        this.wrappedParts = new HashMap();
                    }
                    XMLBridge xmlBridge = p.getInlinedRepeatedElementBridge();
                    if (xmlBridge == null) {
                        xmlBridge = p.getXMLBridge();
                    }
                    this.wrappedParts.put(p.getName(), new ResponseBuilder.WrappedPartBuilder(xmlBridge, setterFactory.get(p)));
                }
                else {
                    try {
                        tempParts.add(new ResponseBuilder.DocLit.PartBuilder(wp.getOwner().getBindingContext().getElementPropertyAccessor(wrapperType, name.getNamespaceURI(), p.getName().getLocalPart()), setterFactory.get(p)));
                        assert p.getBinding() == ParameterBinding.BODY;
                        continue;
                    }
                    catch (JAXBException e) {
                        throw new WebServiceException(wrapperType + " do not have a property of the name " + name, e);
                    }
                }
            }
            this.parts = (ResponseBuilder.DocLit.PartBuilder[])tempParts.toArray(new ResponseBuilder.DocLit.PartBuilder[tempParts.size()]);
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) throws JAXBException, XMLStreamException {
            if (this.dynamicWrapper) {
                return this.readWrappedResponse(msg, args);
            }
            Object retVal = null;
            if (this.parts.length > 0) {
                if (!msg.hasPayload()) {
                    throw new WebServiceException("No payload. Expecting payload with " + this.wrapperName + " element");
                }
                XMLStreamReader reader = msg.readPayload();
                XMLStreamReaderUtil.verifyTag(reader, this.wrapperName);
                Object wrapperBean = this.wrapper.unmarshal(reader, (msg.getAttachments() != null) ? new AttachmentUnmarshallerImpl(msg.getAttachments()) : null);
                try {
                    for (ResponseBuilder.DocLit.PartBuilder part : this.parts) {
                        Object o = part.readResponse(args, wrapperBean);
                        if (o != null) {
                            assert retVal == null;
                            retVal = o;
                        }
                    }
                }
                catch (DatabindingException e) {
                    throw new WebServiceException(e);
                }
                reader.close();
                XMLStreamReaderFactory.recycle(reader);
            }
            else {
                msg.consume();
            }
            return retVal;
        }
        
        static final class PartBuilder
        {
            private final PropertyAccessor accessor;
            private final ValueSetter setter;
            
            public PartBuilder(PropertyAccessor accessor, ValueSetter setter) {
                this.accessor = accessor;
                this.setter = setter;
                assert accessor != null && setter != null;
                return;
            }
            
            final Object readResponse(Object[] args, Object wrapperBean) {
                Object obj = this.accessor.get(wrapperBean);
                return this.setter.put(obj, args);
            }
        }
    }
    
    public static final class RpcLit extends ResponseBuilder
    {
        public RpcLit(WrapperParameter wp, ValueSetterFactory setterFactory) {
            assert wp.getTypeInfo().type == WrapperComposite.class;
            this.wrapperName = wp.getName();
            this.wrappedParts = new HashMap();
            List<ParameterImpl> children = wp.getWrapperChildren();
            for (ParameterImpl p : children) {
                this.wrappedParts.put(p.getName(), new ResponseBuilder.WrappedPartBuilder(p.getXMLBridge(), setterFactory.get(p)));
                assert p.getBinding() == ParameterBinding.BODY;
                continue;
            }
            return;
        }
        
        @Override
        public Object readResponse(Message msg, Object[] args) throws JAXBException, XMLStreamException {
            return this.readWrappedResponse(msg, args);
        }
    }
}
